=== modified file 'loggerhead/controllers/revision_ui.py'
--- loggerhead/controllers/revision_ui.py	2006-12-11 19:23:51 +0000
+++ loggerhead/controllers/revision_ui.py	2006-12-11 21:51:02 +0000
@@ -53,10 +53,7 @@
             # robey-style 1-line long message
             comment = textwrap.wrap(comment[0])
         
-        added, modified, renamed, removed = h.get_file_lists(revid, previous)
-        diff = None
-        if show_diff:
-            diff = h.diff_revisions(revid, previous)
+        changes = h.diff_revisions(revid, previous)
         
         buttons = [
             ('main', turbogears.url('/changes')),
@@ -73,11 +70,7 @@
             'author': rev.committer,
             'comment': comment,
             'date': datetime.datetime.fromtimestamp(rev.timestamp),
-            'files_added': added,
-            'files_modified': modified,
-            'files_renamed': renamed,
-            'files_removed': removed,
-            'diff': diff,
+            'changes': changes,
             'branch_name': turbogears.config.get('loggerhead.branch_name'),
             'util': util,
             'history': h,

=== modified file 'loggerhead/history.py'
--- loggerhead/history.py	2006-12-11 19:23:51 +0000
+++ loggerhead/history.py	2006-12-11 21:51:02 +0000
@@ -16,6 +16,7 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
 
+import cgi
 import datetime
 import textwrap
 from StringIO import StringIO
@@ -24,6 +25,7 @@
 import bzrlib.branch
 import bzrlib.diff
 import bzrlib.errors
+import bzrlib.textfile
 import bzrlib.tsort
 
 from loggerhead import util
@@ -223,34 +225,28 @@
                    self._full_history[max(0, pos - pagesize)])
         else:
             yield ('>', None, None)
-                
+    
     def diff_revisions(self, revid, otherrevid):
-        new_tree = self._branch.repository.revision_tree(revid)
-        old_tree = self._branch.repository.revision_tree(otherrevid)
-        
-        s = StringIO()
-        bzrlib.diff.show_diff_trees(old_tree, new_tree, s, old_label='', new_label='')
-        lines = s.getvalue().split('\n')
-        for l in lines:
-            if l.startswith('='):
-                yield ('info', str(l))
-            elif l.startswith('-'):
-                yield ('old-file', l)
-            elif l.startswith('+'):
-                yield ('new-file', l)
-            elif l.startswith('@'):
-                yield ('lineno', l)
-            else:
-                yield ('diffline', l)
-    
-    def get_file_lists(self, revid, otherrevid):
-        """
-        return 4 lists: (files added, files modified, files renamed,
-           files removed)
-        
-        each is a list of filenames, except the renames, which is a list of
-        (old-name, new-name) tuples.
-        """
+        """
+        Return a nested data structure containing the changes between two
+        revisions::
+        
+            added: list(filename),
+            renamed: list((old_filename, new_filename)),
+            deleted: list(filename),
+            modified: list(
+                filename: str,
+                chunks: list(
+                    diff: list(
+                        old_lineno: int,
+                        new_lineno: int,
+                        type: str('context', 'delete', or 'insert'),
+                        line: str,
+                    ),
+                ),
+            )
+        """
+
         new_tree = self._branch.repository.revision_tree(revid)
         old_tree = self._branch.repository.revision_tree(otherrevid)
         delta = new_tree.changes_from(old_tree)
@@ -267,18 +263,74 @@
                 path += '@'
             return path
         
+        def tree_lines(tree, fid):
+            if not fid in tree:
+                return []
+            tree_file = bzrlib.textfile.text_file(tree.get_file(fid))
+            return tree_file.readlines()
+        
+        def fix_line(s):
+            # to avoid using <pre>, we need to turn spaces into nbsp, and kid
+            # doesn't notice u'\xa0', so we have to do it by hand, and escape
+            # html also.  oh well.
+            s = cgi.escape(s)
+            return s.replace(' ', '&nbsp;')
+        
+        def process_diff(diff):
+            chunks = []
+            chunk = None
+            for line in diff.splitlines():
+                if len(line) == 1:
+                    # i think this happens because bazaar gets too aggressive about removing trailing whitespace
+                    line = line + ' '
+                if line.startswith('+++ ') or line.startswith('--- '):
+                    continue
+                if line.startswith('@@ '):
+                    # new chunk
+                    if chunk is not None:
+                        chunks.append(chunk)
+                    chunk = util.Container()
+                    chunk.diff = []
+                    lines = [int(x.split(',')[0][1:]) for x in line.split(' ')[1:3]]
+                    old_lineno = lines[0]
+                    new_lineno = lines[1]
+                elif line.startswith('  '):
+                    chunk.diff.append(util.Container(old_lineno=old_lineno, new_lineno=new_lineno,
+                                                     type='context', line=fix_line(line[2:])))
+                    old_lineno += 1
+                    new_lineno += 1
+                elif line.startswith('+ '):
+                    chunk.diff.append(util.Container(old_lineno=None, new_lineno=new_lineno,
+                                                     type='insert', line=fix_line(line[2:])))
+                    new_lineno += 1
+                elif line.startswith('- '):
+                    chunk.diff.append(util.Container(old_lineno=old_lineno, new_lineno=None,
+                                                     type='delete', line=fix_line(line[2:])))
+                    old_lineno += 1
+                else:
+                    chunk.diff.append(util.Container(type='unknown', line=fix_line(repr(line))))
+            return chunks
+                    
+        def handle_modify(old_path, new_path, fid, kind):
+            old_lines = tree_lines(old_tree, fid)
+            new_lines = tree_lines(new_tree, fid)
+            buffer = StringIO()
+            bzrlib.diff.internal_diff(old_path, old_lines, new_path, new_lines, buffer)
+            diff = buffer.getvalue()
+            modified.append(util.Container(filename=rich_filename(new_path, kind), chunks=process_diff(diff)))
+
         for path, fid, kind in delta.added:
             added.append(rich_filename(path, kind))
         
         for path, fid, kind, text_modified, meta_modified in delta.modified:
-            modified.append(rich_filename(path, kind))
+            handle_modify(path, path, fid, kind)
         
         for oldpath, newpath, fid, kind, text_modified, meta_modified in delta.renamed:
             renamed.append((rich_filename(oldpath, kind), rich_filename(newpath, kind)))
             if meta_modified or text_modified:
-                modified.append(rich_filename(newpath, kind))
+                handle_modify(oldpath, newpath, fid, kind)
         
         for path, fid, kind in delta.removed:
             removed.append(rich_filename(path, kind))
         
-        return added, modified, renamed, removed
+        return util.Container(added=added, renamed=renamed, removed=removed, modified=modified)

=== modified file 'loggerhead/static/css/style.css'
--- loggerhead/static/css/style.css	2006-12-11 19:23:51 +0000
+++ loggerhead/static/css/style.css	2006-12-11 21:51:02 +0000
@@ -267,10 +267,11 @@
     border-top: 1px solid #8cacbb;
 }
 
+
+/* ----- overly complex diff display ----- */
+
 .diff {
-    font-size: 66%;
     margin-bottom: 1em;
-    xxxborder-top: 1px solid #8cacbb;
     padding-top: 10px;
     margin-top: 20px;
     width: 100%;
@@ -278,9 +279,44 @@
     font-family: monospace;
 }
 
-.diff .old-file { color: red; }
-.diff .new-file { color: green; }
-.diff .lineno { color: purple; background-color: #dee7ec; }
+.diff-block,.diff-chunk {
+	border-collapse: collapse;
+    width: 100%;
+}
+
+.diff-block th.filename {
+	font-family: monospace;
+	font-weight: normal;
+	font-size: smaller;
+	padding-left: 1em;
+	padding-top: 1em;
+	border-top: 1px solid #8cacbb;
+}
+
+.diff-chunk {
+    font-size: 66%;
+    margin-top: 1em;
+	margin-bottom: 1em;
+}
+
+.diff-chunk th.lineno {
+	font-weight: normal;
+    border-bottom: 1px solid #8cacbb;
+}
+.diff-chunk .insert {
+	background-color: #90ff90;
+}
+.diff-chunk .delete {
+	background-color: #ffa0a0;
+}
+
+.diff .lineno {
+	color: purple;
+	background-color: #dee7ec; 
+	width: 3em;
+	text-align: right;
+	padding-right: 3px;
+}
 .diff .info {
 	color: blue;
 	background-color: #dee7ec;

=== modified file 'loggerhead/templates/changelog.kid'
--- loggerhead/templates/changelog.kid	2006-12-11 19:23:51 +0000
+++ loggerhead/templates/changelog.kid	2006-12-11 21:51:02 +0000
@@ -25,8 +25,6 @@
     </table>
 </span>
 
-<!-- #prevpage# -->
-
 <div class="log-entries">
     <table>
         <col class="header" />
@@ -48,7 +46,7 @@
                 </span></td>
             </tr>
             
-            <!-- for multiple-parents: -->
+            <!-- !for multiple-parents: -->
             <span py:if="len(entry.parents) > 1">
                 <span py:for="parent in entry.parents">
                     <tr>
@@ -71,7 +69,5 @@
     </table>
 </div>
 
-<!-- #nextpage# -->
-
 </body>
 </html>

=== modified file 'loggerhead/templates/master.kid'
--- loggerhead/templates/master.kid	2006-12-11 19:23:51 +0000
+++ loggerhead/templates/master.kid	2006-12-11 21:51:02 +0000
@@ -10,9 +10,9 @@
 @import "/static/css/style.css";
     </style>
 
-<!-- define common navbar -->
+<!-- !define common navbar -->
 <span py:def="navbar()">
-    <!-- requires: ${revid}, ${buttons}, ${pagesize} -->
+    <!-- !requires: ${revid}, ${buttons}, ${pagesize} -->
     <div class="navbar">
         <div class="bar">
             <table>

=== modified file 'loggerhead/templates/revision.kid'
--- loggerhead/templates/revision.kid	2006-12-11 19:23:51 +0000
+++ loggerhead/templates/revision.kid	2006-12-11 21:51:02 +0000
@@ -47,31 +47,40 @@
             <td class="description"> <span py:for="line in comment">${line} <br /></span> </td>
         </tr>
         <tr class="divider"> <th></th> <td></td> </tr>
-        <tr py:if="files_added">
+        <tr py:if="changes.added">
             <th class="files"> files added: </th>
-            <td class="files"> <span py:for="filename in files_added">${filename} <br /></span> </td>
+            <td class="files"> <span py:for="filename in changes.added">${filename} <br /></span> </td>
         </tr>
-        <tr py:if="files_removed">
+        <tr py:if="changes.removed">
             <th class="files"> files removed: </th>
-            <td class="files"> <span py:for="filename in files_removed">${filename} <br /></span> </td>
+            <td class="files"> <span py:for="filename in changes.removed">${filename} <br /></span> </td>
         </tr>
-        <tr py:if="files_renamed">
+        <tr py:if="changes.renamed">
             <th class="files"> files renamed: </th>
-            <td class="files"> <span py:for="old_filename, new_filename in files_renamed">${old_filename} => ${new_filename}<br /></span> </td>
+            <td class="files"> <span py:for="old_filename, new_filename in changes.renamed">${old_filename} => ${new_filename}<br /></span> </td>
         </tr>
-        <tr py:if="files_modified">
+        <tr py:if="changes.modified">
             <th class="files"> files modified: </th>
-            <td class="files"> <span py:for="filename in files_modified">${filename} <br /></span> </td>
+            <td class="files"> <span py:for="item in changes.modified">${item.filename} <br /></span> </td>
         </tr>
     </table>
 </div>
 
-<div class="diff" py:if="diff">
-    <div py:for="css_class, line in diff" class="${css_class}"> ${line} </div>
+<div class="diff" py:if="changes.modified">
+    <table py:for="item in changes.modified" class="diff-block">
+        <tr><th class="filename"> ${item.filename} </th></tr>
+        <tr><td>
+            <table py:for="chunk in item.chunks" class="diff-chunk">
+                <tr> <th class="lineno">old</th> <th class="lineno">new</th> <th></th> </tr>
+                <tr py:for="line in chunk.diff">
+                    <td class="lineno">${line.old_lineno}</td>
+                    <td class="lineno">${line.new_lineno}</td>
+                    <td class="${line.type}">${XML(line.line)}</td>
+                </tr>
+            </table>
+        </td></tr>
+    </table>
 </div>
-<span py:if="not diff">
-    <a href="${tg.url([ '/revision', revid ], show_diff=1)}"> (show diff) </a>
-</span>
 
 </body>
 </html>

=== modified file 'loggerhead/util.py'
--- loggerhead/util.py	2006-12-11 06:44:19 +0000
+++ loggerhead/util.py	2006-12-11 21:51:02 +0000
@@ -57,6 +57,15 @@
                 setattr(self, key, value)
         for key, value in kw.iteritems():
             setattr(self, key, value)
+    
+    def __repr__(self):
+        out = '{ '
+        for key, value in self.__dict__.iteritems():
+            if key.startswith('_') or (getattr(self.__dict__[key], '__call__', None) is not None):
+                continue
+            out += '%r => %r, ' % (key, value)
+        out += '}'
+        return out
 
 
 def clean_revid(revid):

